﻿Option Strict On
Option Explicit On

Imports System.IO
Imports System.Runtime.InteropServices

Public Class ConvToMp3

#Region "Implements"
    Implements IDisposable
#End Region

#Region "APIs"

    ' ----==== Mfplat.dll ====----
    <DllImport("Mfplat.dll", EntryPoint:="MFStartup")>
    <PreserveSig> Private Shared Function MFStartup(<[In]> Version As Integer,
                                                    <[In]> dwFlags As Integer) As Integer
    End Function

    <DllImport("Mfplat.dll", EntryPoint:="MFShutdown")>
    <PreserveSig> Private Shared Function MFShutdown() As Integer
    End Function

    <DllImport("Mfplat.dll", EntryPoint:="MFCreateMediaType")>
    <PreserveSig> Private Shared Function MFCreateMediaType(<Out> ByRef ppMFType As IntPtr) As Integer
    End Function

    <DllImport("Mfplat.dll", EntryPoint:="MFCreateAttributes")>
    <PreserveSig> Private Shared Function MFCreateAttributes(<Out> ByRef ppMFAttributes As IntPtr,
                                                             <[In]> cInitialSize As Integer) As Integer
    End Function

    ' ----==== Mfreadwrite.dll ====----
    <DllImport("Mfreadwrite.dll", EntryPoint:="MFCreateSourceReaderFromURL")>
    <PreserveSig> Private Shared Function MFCreateSourceReaderFromURL(<[In], MarshalAs(UnmanagedType.LPWStr)> pwszURL As String,
                                                                      <[In]> pAttributes As IntPtr,
                                                                      <Out> ByRef ppSourceReader As IntPtr) As Integer
    End Function

    <DllImport("Mfreadwrite.dll", EntryPoint:="MFCreateSinkWriterFromURL")>
    <PreserveSig> Private Shared Function MFCreateSinkWriterFromURL(<[In], MarshalAs(UnmanagedType.LPWStr)> pwszOutputURL As String,
                                                                    <[In]> pByteStream As IntPtr,
                                                                    <[In]> pAttributes As IntPtr,
                                                                    <Out> ByRef ppSinkWriter As IntPtr) As Integer
    End Function

    ' ----==== Ole32.dll ====----
    <DllImport("Ole32.dll", EntryPoint:="PropVariantClear")>
    <PreserveSig> Private Shared Function PropVariantClear(<[In]> ByRef pvar As PROPVARIANT) As Integer
    End Function
#End Region

#Region "Class"
    Public Sub New()

        ' die Media Foundation ist noch nicht initialisiert
        If Not m_MFInit Then

            ' Media Foundation initialisiern
            If MFStartup(MF_VERSION_2, MFSTARTUP_FULL) = S_OK Then

                m_MFInit = True

            End If

        End If

    End Sub
#End Region

#Region "Const"
    Private Const S_OK As Integer = &H0

    Private Const MFSTARTUP_FULL As Integer = &H0
    Private Const MF_VERSION_2 As Integer = &H20070

    Private Const SPEAKER_FRONT_LEFT As Integer = &H1
    Private Const SPEAKER_FRONT_RIGHT As Integer = &H2

    Private Const MF_SOURCE_READER_MEDIASOURCE As Long = &HFFFFFFFF
    Private Const MF_SOURCE_READER_FIRST_AUDIO_STREAM As Integer = &HFFFFFFFD

    Private Const MF_PD_DURATION As String = "6c990d33-bb8e-477a-8598-0d5d96fcd88a"

    Private Const MF_MT_SUBTYPE As String = "F7E34C9A-42E8-4714-B74B-CB29D72C35E5"
    Private Const MF_MT_MAJOR_TYPE As String = "48EBA18E-F8C9-4687-BF11-0A74C9F96A8F"

    Private Const MFAudioFormat_MP3 As String = "00000055-0000-0010-8000-00aa00389b71"
    Private Const MFMediaType_Audio As String = "73647561-0000-0010-8000-00AA00389B71"
    Private Const MFAudioFormat_Float As String = "00000003-0000-0010-8000-00aa00389b71"

    Private Const MF_TRANSCODE_CONTAINERTYPE As String = "150ff23f-4abc-478b-ac4f-e1916fba1cca"
    Private Const MFTranscodeContainerType_MP3 As String = "e438b912-83f1-4de6-9e3a-9ffbc6dd24d1"

    Private Const MF_MT_AUDIO_NUM_CHANNELS As String = "37e48bf5-645e-4c5b-89de-ada9e29b696a"
    Private Const MF_MT_AUDIO_CHANNEL_MASK As String = "55fb5765-644a-4caf-8479-938983bb1588"
    Private Const MF_MT_AUDIO_BITS_PER_SAMPLE As String = "f2deb57f-40fa-4764-aa33-ed4f2d1ff669"
    Private Const MF_MT_AUDIO_BLOCK_ALIGNMENT As String = "322de230-9eeb-43bd-ab7a-ff412251541d"
    Private Const MF_MT_ALL_SAMPLES_INDEPENDENT As String = "c9173739-5e56-461c-b713-46fb995cb95f"
    Private Const MF_MT_AUDIO_SAMPLES_PER_SECOND As String = "5faeeae7-0290-4c31-9e8a-c534f68d9dba"
    Private Const MF_MT_AUDIO_AVG_BYTES_PER_SECOND As String = "1aab75c8-cfef-451c-ab95-ac034b8e1731"
#End Region

#Region "Enums"

    Private Enum VTable_Interfaces As Integer

        ' IMFSourceReader
        SetCurrentMediaType = 7
        ReadSample = 9
        GetPresentationAttribute = 12

        ' IMFSinkWriter
        AddStream = 3
        SetInputMediaType = 4
        BeginWriting = 5
        WriteSample = 6
        Finalize = 11

        ' IMFMediaType / IMFAttributes
        SetUINT32 = 21
        SetGUID = 24

        ' IMFSample
        GetSampleDuration = 37

    End Enum

    Private Enum MF_SOURCE_READER_FLAG As Integer
        MF_SOURCE_READERF_ERROR = &H1
        MF_SOURCE_READERF_ENDOFSTREAM = &H2
        MF_SOURCE_READERF_NEWSTREAM = &H4
        MF_SOURCE_READERF_NATIVEMEDIATYPECHANGED = &H10
        MF_SOURCE_READERF_CURRENTMEDIATYPECHANGED = &H20
        MF_SOURCE_READERF_STREAMTICK = &H100
        MF_SOURCE_READERF_ALLEFFECTSREMOVED = &H200
    End Enum
#End Region

#Region "Events"
    Public Event Converted(Percent As Integer, [Error] As Boolean)
#End Region

#Region "Structure"
    Private Structure CArray
        Dim cElems As UInteger
        Dim pElems As IntPtr
    End Structure

    <StructLayout(LayoutKind.Explicit, Size:=16)>
    Private Structure PROPVARIANT
        <FieldOffset(0)> Dim vt As UShort
        <FieldOffset(2)> Dim wReserved1 As UShort
        <FieldOffset(4)> Dim wReserved2 As UShort
        <FieldOffset(6)> Dim wReserved3 As UShort
        <FieldOffset(8)> Dim uintVal As UInteger 'UINT32
        <FieldOffset(8)> Dim ulVal As ULong      'UINT64
        <FieldOffset(8)> Dim dblVal As Double    'DOUBLE
        <FieldOffset(8)> Dim pwszVal As IntPtr   'STRING
        <FieldOffset(8)> Dim punkVal As IntPtr   'IUNKNOWN
        <FieldOffset(8)> Dim cArray As CArray    'BLOB
    End Structure
#End Region

#Region "Variablen"
    Private m_MFInit As Boolean = False
    Private m_DisposedValue As Boolean = False
    Private m_InputFile As String = String.Empty
    Private m_OutputFile As String = String.Empty
#End Region

#Region "Properties"
    Public ReadOnly Property IsInitialized As Boolean
        Get
            Return m_MFInit
        End Get
    End Property

    Public WriteOnly Property InputFile As String
        Set(value As String)
            m_InputFile = value
        End Set
    End Property

    Public WriteOnly Property OutputFile As String
        Set(value As String)
            m_OutputFile = value
        End Set
    End Property
#End Region

#Region "Public Funktions"

    Public Function Convert() As Boolean

        Return ConvertIntern(m_InputFile, m_OutputFile)

    End Function

    Public Function Convert(InputFile As String, OutputFile As String) As Boolean

        Return ConvertIntern(InputFile, OutputFile)

    End Function
#End Region

#Region "Private Functions"
    Private Function ConvertIntern(InputFile As String, OutputFile As String) As Boolean

        Dim bolRet As Boolean = False

        ' ist die Media Foundation initialisiert
        If m_MFInit Then

            ' Strings nicht Null oder Empty
            If Not String.IsNullOrEmpty(InputFile) And Not String.IsNullOrEmpty(OutputFile) Then

                ' die Input-Datei existiert und die Output-Datei hat die Dateiextension .MP3
                If File.Exists(InputFile) And OutputFile.ToLower.EndsWith(".mp3") Then

                    Dim pIMFSourceReader As IntPtr = IntPtr.Zero

                    ' einen SourceReader von der Input-Datei erstellen
                    If CreateSourceReader(InputFile, pIMFSourceReader) Then

                        Dim ulngDuration As ULong
                        Dim tPROPVARIANT As New PROPVARIANT

                        ' Duration der Input-Datei ermitteln
                        If MFSourceReader_GetPresentationAttribute(pIMFSourceReader, MF_SOURCE_READER_MEDIASOURCE,
                                                                   New Guid(MF_PD_DURATION), tPROPVARIANT) Then

                            ' Duration speichern
                            ulngDuration = tPROPVARIANT.ulVal

                            ' PROPVARIANT löschen
                            PropVariantClear(tPROPVARIANT)

                        End If

                        Dim intStreamIndex As Integer
                        Dim pIMFSinkWriter As IntPtr = IntPtr.Zero

                        ' einen SinkWriter für die Output-Datei erstellen
                        If CreateSinkWriter(OutputFile, pIMFSinkWriter, intStreamIndex) Then

                            ' mit dem Schreiben von Daten in den SinkWriter beginnen
                            If MFSinkWriter_BeginWriting(pIMFSinkWriter) Then

                                Dim bolError As Boolean = False
                                Dim bolEndOfStream As Boolean = False

                                Do

                                    Dim lngTimeStamp As Long
                                    Dim intActualStreamIndex As Integer
                                    Dim pIMFSample As IntPtr = IntPtr.Zero
                                    Dim eStreamFlags As MF_SOURCE_READER_FLAG

                                    ' der SourceReader ließt ein Sample von der Input-Datei
                                    If MFSourceReader_ReadSample(pIMFSourceReader, MF_SOURCE_READER_FIRST_AUDIO_STREAM, 0,
                                                                 intActualStreamIndex, eStreamFlags, lngTimeStamp, pIMFSample) Then

                                        ' wenn das Ende des Audiostreams erreicht ist
                                        If CBool(eStreamFlags And MF_SOURCE_READER_FLAG.MF_SOURCE_READERF_ENDOFSTREAM) Then

                                            ' Flag setzen
                                            bolEndOfStream = True

                                        Else
                                            ' wenn das Ende des Audiostreams noch nicht erreicht ist

                                            ' ist ein Sample vorhanden
                                            If pIMFSample <> IntPtr.Zero Then

                                                ' der SinkWriter schreibt das Sample in die Output-Datei
                                                If MFSinkWriter_WriteSample(pIMFSinkWriter, intStreamIndex, pIMFSample) Then

                                                    Dim lngSampleDuration As Long

                                                    ' Duration des Sample ermitteln
                                                    If MFSample_GetSampleDuration(pIMFSample, lngSampleDuration) Then

                                                        ' ist eine Duration von der Input-Datei vorhanden
                                                        If ulngDuration > 0 Then

                                                            ' Fortschritt der Konvertierung in Prozent umrechnen
                                                            Dim intPercent As Integer = CInt(((CULng(lngTimeStamp) + CULng(lngSampleDuration)) / ulngDuration) * 100UL)

                                                            ' Event auslösen
                                                            RaiseEvent Converted(intPercent, False)

                                                        End If

                                                    End If

                                                Else
                                                    ' es gab einen Fehler beim schreiben des Sample

                                                    ' Flags setzen
                                                    bolError = True
                                                    bolEndOfStream = True

                                                    ' Event auslösen
                                                    RaiseEvent Converted(100, True)

                                                End If

                                            End If

                                        End If

                                        ' ist ein Sample vorhanden, dann dieses löschen
                                        If pIMFSample <> IntPtr.Zero Then Marshal.Release(pIMFSample)

                                    End If

                                    ' Schleife beenden wenn das Ende das Streams erreicht wurde
                                    ' oder es einen Fehler beim schreiben eines Sample gab
                                Loop Until bolEndOfStream = True

                                ' schreiben des SinkWriters beenden und finalisieren
                                If MFSinkWriter_Finalize(pIMFSinkWriter) Then

                                    ' wenn es bis hierher keinen Fehler gab
                                    If Not bolError Then bolRet = True

                                End If

                            End If

                            ' SinkWriter löschen
                            Marshal.Release(pIMFSinkWriter)

                        End If

                        ' SourceReader löschen
                        Marshal.Release(pIMFSourceReader)

                    End If

                End If

            End If

        End If

        Return bolRet

    End Function

    Private Function CreateSourceReader(sURL As String,
                                        ByRef pSourceReader As IntPtr) As Boolean

        Dim bolRet As Boolean = False

        ' einen SourceReader von einer Datei erstellen
        If MFCreateSourceReaderFromURL(sURL, IntPtr.Zero, pSourceReader) = S_OK Then

            Dim pTransferAudioType As IntPtr = IntPtr.Zero

            ' ein Audio-Übertragungsformat erstellen
            ' muss das gleiche sein das für den SinkWriter verwendet wird
            If CreateTransferAudioType(pTransferAudioType) Then

                ' Übertragungsformat für den Audio-Stream festlegen
                If MFSourceReader_SetCurrentMediaType(pSourceReader, MF_SOURCE_READER_FIRST_AUDIO_STREAM,
                                                      IntPtr.Zero, pTransferAudioType) Then

                    bolRet = True

                End If

                ' Übertragungsformat löschen
                Marshal.Release(pTransferAudioType)

            End If

        End If

        Return bolRet

    End Function

    Private Function CreateSinkWriter(sURL As String,
                                      ByRef pSinkWriter As IntPtr,
                                      ByRef intStreamIndex As Integer) As Boolean

        Dim bolRet As Boolean = False
        Dim pAttributes As IntPtr = IntPtr.Zero

        ' ein IMFAttribute erstellen
        If MFCreateAttributes(pAttributes, 0) = S_OK Then

            ' Containerformat, hier MP3, für die Output-Datei festlegen
            If MFAttributes_SetGUID(pAttributes, New Guid(MF_TRANSCODE_CONTAINERTYPE),
                                    New Guid(MFTranscodeContainerType_MP3)) Then

                ' einen SinkWriter für die Output-Datei erstellen
                If MFCreateSinkWriterFromURL(sURL, IntPtr.Zero, pAttributes, pSinkWriter) = S_OK Then

                    Dim pMp3AudioType As IntPtr = IntPtr.Zero

                    ' ein Ausgabeformat für MP3 erstellen
                    If CreateMp3AudioType(pMp3AudioType) Then

                        ' Ausgabeformat für den SinkWriter festlegen
                        If MFSinkWriter_AddStream(pSinkWriter, pMp3AudioType, intStreamIndex) Then

                            Dim pTransferAudioType As IntPtr = IntPtr.Zero

                            ' ein Audio-Übertragungsformat erstellen
                            ' muss das gleiche sein das für den SourceReader verwendet wird
                            If CreateTransferAudioType(pTransferAudioType) Then

                                ' Übertragungsformat für den Audio-Stream festlegen
                                If MFSinkWriter_SetInputMediaType(pSinkWriter, intStreamIndex,
                                                                  pTransferAudioType, IntPtr.Zero) Then

                                    bolRet = True

                                End If

                                ' Übertragungsformat löschen
                                Marshal.Release(pTransferAudioType)

                            End If

                        End If

                        ' Ausgabeformat löschen
                        Marshal.Release(pMp3AudioType)

                    End If

                End If

            End If

            ' IMFAttribute löschen
            Marshal.Release(pAttributes)

        End If

        Return bolRet

    End Function

    Private Function CreateMp3AudioType(ByRef pMp3AudioType As IntPtr) As Boolean

        Dim bolRet As Boolean = False

        ' ein IMFMediaType erstellen
        If MFCreateMediaType(pMp3AudioType) = S_OK Then

            ' der MediaType ist ein Audiostream
            If MFAttributes_SetGUID(pMp3AudioType,
                                    New Guid(MF_MT_MAJOR_TYPE),
                                    New Guid(MFMediaType_Audio)) Then

                ' das Audioformat ist MP3
                If MFAttributes_SetGUID(pMp3AudioType,
                                        New Guid(MF_MT_SUBTYPE),
                                        New Guid(MFAudioFormat_MP3)) Then

                    ' 44,1kHz
                    If MFAttributes_SetUINT32(pMp3AudioType,
                                              New Guid(MF_MT_AUDIO_SAMPLES_PER_SECOND),
                                              44100) Then
                        ' Stereo
                        If MFAttributes_SetUINT32(pMp3AudioType,
                                                  New Guid(MF_MT_AUDIO_NUM_CHANNELS),
                                                  2) Then

                            ' 320kb/s
                            If MFAttributes_SetUINT32(pMp3AudioType,
                                                      New Guid(MF_MT_AUDIO_AVG_BYTES_PER_SECOND),
                                                      40000) Then

                                ' BlockAlignment
                                If MFAttributes_SetUINT32(pMp3AudioType,
                                                          New Guid(MF_MT_AUDIO_BLOCK_ALIGNMENT),
                                                          1) Then

                                    bolRet = True

                                End If

                            End If

                        End If

                    End If

                End If

            End If

        End If

        Return bolRet

    End Function

    Private Function CreateTransferAudioType(ByRef pTransferAudioType As IntPtr) As Boolean

        Dim bolRet As Boolean = False

        ' ein IMFMediaType erstellen
        If MFCreateMediaType(pTransferAudioType) = S_OK Then

            ' der MediaType ist ein Audiostream
            If MFAttributes_SetGUID(pTransferAudioType,
                                    New Guid(MF_MT_MAJOR_TYPE),
                                    New Guid(MFMediaType_Audio)) Then

                ' diverse Eigenschaften festlegen. Anhand der MF_MT_AUDIO_-GUIDs
                ' ist ja ersichtlich was hier festgelegt wird.
                If MFAttributes_SetGUID(pTransferAudioType,
                                        New Guid(MF_MT_SUBTYPE),
                                        New Guid(MFAudioFormat_Float)) Then

                    If MFAttributes_SetUINT32(pTransferAudioType,
                                              New Guid(MF_MT_AUDIO_SAMPLES_PER_SECOND),
                                              48000) Then

                        If MFAttributes_SetUINT32(pTransferAudioType,
                                                  New Guid(MF_MT_AUDIO_BITS_PER_SAMPLE),
                                                  32) Then

                            If MFAttributes_SetUINT32(pTransferAudioType,
                                                      New Guid(MF_MT_AUDIO_NUM_CHANNELS),
                                                      2) Then

                                If MFAttributes_SetUINT32(pTransferAudioType,
                                                          New Guid(MF_MT_AUDIO_AVG_BYTES_PER_SECOND),
                                                          384000) Then

                                    If MFAttributes_SetUINT32(pTransferAudioType,
                                                              New Guid(MF_MT_AUDIO_BLOCK_ALIGNMENT),
                                                              8) Then

                                        If MFAttributes_SetUINT32(pTransferAudioType,
                                                                  New Guid(MF_MT_ALL_SAMPLES_INDEPENDENT),
                                                                  1) Then

                                            If MFAttributes_SetUINT32(pTransferAudioType,
                                                                      New Guid(MF_MT_AUDIO_CHANNEL_MASK),
                                                                      SPEAKER_FRONT_LEFT Or SPEAKER_FRONT_RIGHT) Then

                                                bolRet = True

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

            End If

        End If

        Return bolRet

    End Function
#End Region

#Region "Interface Delegates"
    Private Delegate Function IMFAttributes_SetGUID(<[In]> this As IntPtr,
                                                    <[In]> ByRef guidKey As Guid,
                                                    <[In]> ByRef guidValue As Guid) As Integer

    Private Delegate Function IMFAttributes_SetUINT32(<[In]> this As IntPtr,
                                                      <[In]> ByRef guidKey As Guid,
                                                      <[In]> unValue As UInteger) As Integer

    Private Delegate Function IMFSourceReader_ReadSample(<[In]> this As IntPtr,
                                                         <[In]> dwStreamIndex As Integer,
                                                         <[In]> dwControlFlags As Integer,
                                                         <Out> ByRef pdwActualStreamIndex As Integer,
                                                         <Out> ByRef pdwStreamFlags As MF_SOURCE_READER_FLAG,
                                                         <Out> ByRef pllTimestamp As Long,
                                                         <Out> ByRef ppSample As IntPtr) As Integer

    Private Delegate Function IMFSourceReader_SetCurrentMediaType(<[In]> this As IntPtr,
                                                                  <[In]> dwStreamIndex As Integer,
                                                                  <[In]> pdwReserved As IntPtr,
                                                                  <[In]> pMediaType As IntPtr) As Integer

    Private Delegate Function IMFSourceReader_GetPresentationAttribute(<[In]> this As IntPtr,
                                                                       <[In]> dwStreamIndex As Integer,
                                                                       <[In]> ByRef guidAttribute As Guid,
                                                                       <Out> ByRef pvarAttribute As PROPVARIANT) As Integer

    Private Delegate Function IMFSinkWriter_Finalize(<[In]> this As IntPtr) As Integer

    Private Delegate Function IMFSinkWriter_AddStream(<[In]> this As IntPtr,
                                                      <[In]> pTargetMediaType As IntPtr,
                                                      <Out> ByRef pdwStreamIndex As Integer) As Integer

    Private Delegate Function IMFSinkWriter_WriteSample(<[In]> this As IntPtr,
                                                        <[In]> dwStreamIndex As Integer,
                                                        <[In]> pSample As IntPtr) As Integer

    Private Delegate Function IMFSinkWriter_BeginWriting(<[In]> this As IntPtr) As Integer

    Private Delegate Function IMFSinkWriter_SetInputMediaType(<[In]> this As IntPtr,
                                                              <[In]> dwStreamIndex As Integer,
                                                              <[In]> pInputMediaType As IntPtr,
                                                              <[In]> pEncodingParameters As IntPtr) As Integer

    Private Delegate Function IMFSample_GetSampleDuration(<[In]> this As IntPtr,
                                                          <Out> ByRef phnsSampleDuration As Long) As Integer
#End Region

#Region "Interface Functions"
    Private Function MFAttributes_SetGUID(this As IntPtr,
                                          guidKey As Guid,
                                          guidValue As Guid) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.SetGUID)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFAttributes_SetGUID = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                 GetType(IMFAttributes_SetGUID)),
                                                                 IMFAttributes_SetGUID)

                If Method(this, guidKey, guidValue) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFAttributes_SetUINT32(this As IntPtr,
                                            guidKey As Guid,
                                            unValue As UInteger) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.SetUINT32)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFAttributes_SetUINT32 = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                   GetType(IMFAttributes_SetUINT32)),
                                                                   IMFAttributes_SetUINT32)

                If Method(this, guidKey, unValue) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSourceReader_SetCurrentMediaType(this As IntPtr,
                                                        dwStreamIndex As Integer,
                                                        pdwReserved As IntPtr,
                                                        pMediaType As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.SetCurrentMediaType)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSourceReader_SetCurrentMediaType = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                               GetType(IMFSourceReader_SetCurrentMediaType)),
                                                                               IMFSourceReader_SetCurrentMediaType)

                If Method(this, dwStreamIndex, pdwReserved, pMediaType) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSourceReader_ReadSample(this As IntPtr,
                                               dwStreamIndex As Integer,
                                               dwControlFlags As Integer,
                                               ByRef pdwActualStreamIndex As Integer,
                                               ByRef pdwStreamFlags As MF_SOURCE_READER_FLAG,
                                               ByRef pllTimestamp As Long,
                                               ByRef ppSample As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.ReadSample)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSourceReader_ReadSample = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                      GetType(IMFSourceReader_ReadSample)),
                                                                      IMFSourceReader_ReadSample)

                If Method(this, dwStreamIndex, dwControlFlags, pdwActualStreamIndex,
                          pdwStreamFlags, pllTimestamp, ppSample) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSourceReader_GetPresentationAttribute(this As IntPtr,
                                                             dwStreamIndex As Integer,
                                                             guidAttribute As Guid,
                                                             ByRef pvarAttribute As PROPVARIANT) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.GetPresentationAttribute)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSourceReader_GetPresentationAttribute = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                                    GetType(IMFSourceReader_GetPresentationAttribute)),
                                                                                    IMFSourceReader_GetPresentationAttribute)

                If Method(this, dwStreamIndex, guidAttribute, pvarAttribute) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSinkWriter_Finalize(this As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.Finalize)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSinkWriter_Finalize = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                  GetType(IMFSinkWriter_Finalize)),
                                                                  IMFSinkWriter_Finalize)

                If Method(this) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSinkWriter_AddStream(this As IntPtr,
                                            pTargetMediaType As IntPtr,
                                            ByRef pdwStreamIndex As Integer) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.AddStream)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSinkWriter_AddStream = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                   GetType(IMFSinkWriter_AddStream)),
                                                                   IMFSinkWriter_AddStream)

                If Method(this, pTargetMediaType, pdwStreamIndex) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSinkWriter_BeginWriting(this As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.BeginWriting)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSinkWriter_BeginWriting = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                      GetType(IMFSinkWriter_BeginWriting)),
                                                                      IMFSinkWriter_BeginWriting)

                If Method(this) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSinkWriter_WriteSample(this As IntPtr,
                                              dwStreamIndex As Integer,
                                              pSample As IntPtr) As Boolean
        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.WriteSample)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSinkWriter_WriteSample = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                     GetType(IMFSinkWriter_WriteSample)),
                                                                     IMFSinkWriter_WriteSample)

                If Method(this, dwStreamIndex, pSample) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSinkWriter_SetInputMediaType(this As IntPtr,
                                                    dwStreamIndex As Integer,
                                                    pInputMediaType As IntPtr,
                                                    pEncodingParameters As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.SetInputMediaType)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSinkWriter_SetInputMediaType = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                           GetType(IMFSinkWriter_SetInputMediaType)),
                                                                           IMFSinkWriter_SetInputMediaType)

                If Method(this, dwStreamIndex, pInputMediaType, pEncodingParameters) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function MFSample_GetSampleDuration(this As IntPtr,
                                                ByRef phnsSampleDuration As Long) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.GetSampleDuration)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFSample_GetSampleDuration = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                       GetType(IMFSample_GetSampleDuration)),
                                                                       IMFSample_GetSampleDuration)

                If Method(this, phnsSampleDuration) = S_OK Then

                    Return True

                End If

            End If

        End If

        Return False

    End Function

    Private Function GetMethodPtr(pInterface As IntPtr,
                                  intMethodNumber As Integer) As IntPtr

        Dim pMethodPtr As IntPtr = IntPtr.Zero

        If pInterface <> IntPtr.Zero And intMethodNumber >= 0 Then

            Dim pMethod As IntPtr() = New IntPtr(0) {}

            Marshal.Copy(IntPtr.Add(Marshal.ReadIntPtr(pInterface),
                         IntPtr.Size * intMethodNumber), pMethod, 0, 1)

            pMethodPtr = pMethod(0)

        End If

        Return pMethodPtr

    End Function
#End Region

#Region "Implements IDisposable"
    Protected Overridable Sub Dispose(disposing As Boolean)

        If Not m_DisposedValue Then

            If disposing Then

                ' die Media Foundation ist initialisiert
                If m_MFInit Then

                    ' Media Foundation ist noch nicht initialisiert
                    If MFShutdown() = S_OK Then m_MFInit = False

                End If

            End If

            m_DisposedValue = True

        End If

    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose

        Dispose(disposing:=True)

        GC.SuppressFinalize(Me)

    End Sub
#End Region

End Class
